#include "Document.h"
#include <assert.h>
#include <nlp_config.h>
#include <nlp_filesystem.h>
#include <nlp_macros.h>
#include "Vocabulary.h"
#include "Sample.h"
#include "Learner.h"


IdToDocument_map_t			Document::map_IdToDocument;
StartingStateToIds_map_t	Document::map_StartingStateToIds;
Document_vec_t				Document::vec_Documents;
String_dq_t					Document::dq_StartingStates;


//													
std::ostream& operator<< (std::ostream& _rStream, const DocumentType_e _eType)
{
	if (dt_training == _eType)
		_rStream << "(training)";
	else if (dt_development == _eType)
		_rStream << "(development)";
	else if (dt_test == _eType)
		_rStream << "(test)";
	else
		_rStream << "(unknown-document-type)";
	return _rStream;
}


//													
std::ostream& operator<< (std::ostream& _rStream, const Sentence& _rSentence)
{
	String_vec_t::iterator	ite = ((Sentence&)_rSentence).vec_Words.begin ();
	String_vec_t::iterator	ite_end = ((Sentence&)_rSentence).vec_Words.end ();
	if (ite_end == ite)
		return _rStream;

	_rStream << *ite;
	++ ite;
	for (; ite != ite_end; ++ ite)
		_rStream << ' ' << *ite;

	return _rStream;
}




//													
//													
TraceInfo::~TraceInfo (void)
{
	ITERATE (AnnotatedAction_dq_t, dq_Actions, ite)
		delete *ite;
	dq_Actions.clear ();
}


//													
void TraceInfo::RememberActions (EpisodeStep_dq_t& _rdqSteps, int _iEnd)
{
	// delete old actions if present ...
	ITERATE (AnnotatedAction_dq_t, dq_Actions, ite)
		delete *ite;
	dq_Actions.clear ();

	// remember new actions ...			
	for (int i = 0; i < _iEnd; ++ i)
	{
		EpisodeStep* pStep = _rdqSteps [i];
		RLAction& rAction = pStep->o_RLAction;
		RLState& rState = pStep->o_RLState;

		AnnotatedAction* pAction = new AnnotatedAction;
		pAction->i_CommandWord = rAction.i_SelectedCommandWordLocation;
		pAction->i_ObjectWord = rAction.i_SelectedObjectWordLocation;
		pAction->i_Parameter = rAction.i_SelectedParameterWordLocation;
		pAction->s_Command = Command::Get (rAction.i_SelectedCommandIndex);

		UIObject* pObject = rState.GetObject (rAction.i_SelectedObjectIndex);
		pAction->s_ObjectName = pObject->s_CleanName;
		pAction->s_ObjectType = ObjectClass::Get (pObject->i_ObjectClass);
		pAction->s_TemporallyUniqueObjectId = pObject->s_TemporallyUniqueId;

		dq_Actions.push_back (pAction);
	}
}




//													
//													
Document::Document (void)
{
	e_Type = dt_training;
	p_StartingEnhancedState = NULL;
	p_TrainingEnhancedState = new EnhancedState;
	p_TestEnhancedState = new EnhancedState;
	i_WordCount = 0;
	parr_LocationToWord = NULL;
}


//													
Document::~Document (void)
{
	if (NULL != p_TrainingEnhancedState)
		delete p_TrainingEnhancedState;
	delete p_TestEnhancedState;
}


//													
void Document::SetType (DocumentType_e _eType)
{
	e_Type = _eType;
	if (dt_training == e_Type)
		return;

	delete p_TrainingEnhancedState;
	p_TrainingEnhancedState = NULL;
}


//													
void Document::RememberStates (EpisodeStep_dq_t& _dqSteps)
{
	//									
	for (size_t i = 1; i < _dqSteps.size (); ++ i)
	{
		EpisodeStep* pStep = _dqSteps [i];
		EnhancedState* pEnhancedState = pStep->o_RLState.p_EnhancedState;

		double dReward = pStep->o_RLState.GetReward ();
		bool bDeadend = ((-1 * Learner::d_NegativeRewardMultiplier == dReward) ||
						(true == isnan (dReward)) ||
						(ft_deadend == pStep->e_Feedback) ||
						(ft_state_action_error == pStep->e_Feedback) ||
						(ft_comm_error == pStep->e_Feedback));

		VisitedStates_map_t::iterator	ite;
		ite = map_VisitedStates.find (pEnhancedState);
		if (map_VisitedStates.end () == ite)
		{
			if (false == bDeadend)
			{
				pair <VisitedStates_map_t::iterator, bool> pairInsert;
				pairInsert = map_VisitedStates.insert (make_pair (pEnhancedState, TraceInfo ()));
				TraceInfo& rInfo = pairInsert.first->second;
				rInfo.f_Reward = dReward;
				rInfo.RememberActions (_dqSteps, i);
			}
		}
		else
		{
			if (false == bDeadend)
			{
				TraceInfo& rInfo = ite->second;
				if (rInfo.f_Reward < dReward)
				{
					rInfo.f_Reward = dReward;
					rInfo.RememberActions (_dqSteps, i);
				}
			}
			else
				map_VisitedStates.erase (ite);
		}
	}
}


//													
AnnotatedAction_dq_t* Document::PickStartState (void)
{
	int iStartStates = map_VisitedStates.size ();
	if (0 == iStartStates)
		return NULL;
	cout << "   [**] picking start state from " << iStartStates << endl;
	int iSample = Sample::SampleUniformCategorical (iStartStates);

	int iIndex = 0;
	ITERATE (VisitedStates_map_t, map_VisitedStates, ite)
	{
		if (iSample == iIndex)
		{
			TraceInfo& rInfo = ite->second;
			return &rInfo.dq_Actions;
		}
		++ iIndex;
	}

	return NULL;
}






//													
//													
void Document::ComputeWordIndices (void)
{
	i_WordCount = 0;
	ITERATE (Sentence_vec_t, vec_Sentences, ite)
		i_WordCount += (*ite)->vec_Words.size ();

	parr_LocationToWord = new WordLocation [i_WordCount];

	int iLocation = 0;
	int iSentence = 0;
	ITERATE (Sentence_vec_t, vec_Sentences, ite)
	{
		Sentence* pSentence = *ite;
		int iSentenceWords = pSentence->vec_WordIndices.size ();
		for (int iWord = 0; iWord < iSentenceWords; ++ iWord)
		{
			parr_LocationToWord [iLocation].i_SentenceIndex = iSentence;
			parr_LocationToWord [iLocation].i_WordIndex = iWord;
			++ iLocation;
		}
		++ iSentence;
	}

	if (i_WordCount >= 128)
	{
		cout << "Document [" << s_Id << "] has more than 128 words. \n"
			 << "This codebase needs to be modified to support that." << endl;
		abort ();
	}
}


//													
void Document::SplitToWords (String _sLine, String_dq_t* _pdqWords)
{
	_sLine.LowerCase ();
	_sLine.Strip ();
	_sLine.Replace (",", " , ");
	_sLine.ReplacePattern ("[ \t]+", " ");

	String sCleanLine;

	size_t iLast = 0;
	size_t iStart = _sLine.Find ("<b>");
	while (string::npos != iStart)
	{
		size_t iEnd = _sLine.Find ("</b>", iStart);
		assert (string::npos != iEnd);

		sCleanLine << _sLine.substr (iLast, iStart - iLast);
		iLast = iEnd;

		for (size_t i = iStart; i < iEnd; ++ i)
		{
			char ch = _sLine [i];
			sCleanLine << ((' ' == ch)? (char)'\x01' : ch);
		}

		iStart = _sLine.Find ("<b>", iEnd);
	}

	sCleanLine << _sLine.substr (iLast);
	sCleanLine.Replace ("</b>" , " ");
	sCleanLine.Replace ("<b>" , " ");

	sCleanLine.Split (*_pdqWords);
	ITERATE (String_dq_t, (*_pdqWords), ite)
	{
		(*ite).Replace ("\x01", " ");
		(*ite).Replace (" ,", ",");
		(*ite).Replace (",  ", ", ");
	}
}



//													
bool Document::Load (void)
{
	cout << "Loading documents ..." << endl;
	if (false == Document::LoadDocuments ())
		return false;
	if (false == Document::LoadHeldoutIds ())
		return false;
	return true;
}


//													
bool Document::LoadDocuments (void)
{
	Clear ();

	// load documents ...										
	String_dq_t	dqLines;
	if (false == File::ReadLines ((config)"documents_file", dqLines))
		return false;

	Document* pDocument = NULL;
	ITERATE (String_dq_t, dqLines, ite)
	{
		(*ite).Strip ();
		if (true == (*ite).StartsWith ("#"))
			continue;

		// document id						
		if (true == (*ite).StartsWith ("!"))
		{
			assert (NULL == pDocument);
			pDocument = new Document;
			pDocument->s_Id = (*ite).substr (1);
			pDocument->s_Id.Strip ();
			continue;
		}

		// required starting state			
		if (true == (*ite).StartsWith ("@"))
		{
			assert (NULL != pDocument);
			pDocument->s_RequiredStartingState = (*ite).substr (1);
			pDocument->s_RequiredStartingState.Strip ();
			continue;
		}

		// end of document					
		if ("" == *ite)
		{
			if (NULL == pDocument)
				continue;
			map_IdToDocument.insert (make_pair (pDocument->s_Id, pDocument));
			pDocument = NULL;
			continue;
		}

		String_dq_t	dqWords;
		Document::SplitToWords (*ite, &dqWords);

		Sentence* pSentence = new Sentence;
		pDocument->vec_Sentences.push_back (pSentence);
		pSentence->vec_Words.resize (dqWords.size ());
		pSentence->vec_WordIndices.resize (dqWords.size ());

		for (size_t i = 0; i < dqWords.size (); ++ i)
		{
			String& sWord = dqWords [i];
			pSentence->vec_Words [i] = sWord;
			pSentence->vec_WordIndices [i] = (word_t) sWord;
		}
		pSentence->i_Length = pSentence->vec_Words.size ();
	}
	if (NULL != pDocument)
		map_IdToDocument.insert (make_pair (pDocument->s_Id, pDocument));


	vec_Documents.reserve (map_IdToDocument.size ());

	// Construct required-starting-state to document-ids map	
	ITERATE (IdToDocument_map_t, map_IdToDocument, ite)
	{
		Document* pDocument = ite->second;
		pDocument->ComputeWordIndices ();

		vec_Documents.push_back (pDocument);

		String_dq_t* pdqIds = GetIdsForStartingState (pDocument->s_RequiredStartingState);
		if (NULL == pdqIds)
		{
			pdqIds = new String_dq_t;
			map_StartingStateToIds.insert (make_pair (pDocument->s_RequiredStartingState, pdqIds));
			dq_StartingStates.push_back (pDocument->s_RequiredStartingState);
		}
		pdqIds->push_back (pDocument->s_Id);
	}

	cout << "   - documents : " << map_IdToDocument.size () << endl;
	return true;
}


//													
bool Document::LoadHeldoutIds (void)
{
	String_dq_t	dqHeldoutIds;
	if (false == File::ReadLines ((config)"heldout_id_file", dqHeldoutIds))
		return false;

	long lTrainingDocs = map_IdToDocument.size ();
	long lDevDocs = 0;
	long lTestDocs = 0;

	bool bOk = true;
	ITERATE (String_dq_t, dqHeldoutIds, ite)
	{
		String_dq_t dqValues;
		(*ite).Split (dqValues);

		Document* pDocument = Document::Find (dqValues [0]);
		if (NULL == pDocument)
		{
			cout << "[ERROR]  Document [id=" << *ite << "] marked as heldout, not found."
				 << endl;
			continue;
		}
		if ("development" == dqValues [1])
		{
			pDocument->SetType (dt_development);
			++ lDevDocs;
		}
		else if ("test" == dqValues [1])
		{
			pDocument->SetType (dt_test);
			++ lTestDocs;
		}
		else
		{
			cout << "[ERROR]  Unknown document type (" << dqValues [1]
				 << ") for document [id=" << dqValues [0] << endl;
			bOk = false;
		}
	}

	cout << "   - training : " << lTrainingDocs - lTestDocs - lDevDocs << endl;
	cout << "   - dev      : " << lDevDocs << endl;
	cout << "   - test     : " << lTestDocs << endl;
	return bOk;
}


//													
void Document::Clear (void)
{
	ITERATE (IdToDocument_map_t, map_IdToDocument, ite)
		delete ite->second;
	map_IdToDocument.clear ();

	ITERATE (StartingStateToIds_map_t, map_StartingStateToIds, ite)
		delete ite->second;
	map_StartingStateToIds.clear ();
}


//													
void Document::ForgetTestStateGraph (void)
{
	ITERATE (IdToDocument_map_t, map_IdToDocument, ite)
	{
		Document* pDoc = ite->second;
		delete pDoc->p_TestEnhancedState;
		pDoc->p_TestEnhancedState = new EnhancedState;
	}
}


//													
void Document::SwitchToTrainingStateGraph (void)
{
	ITERATE (IdToDocument_map_t, map_IdToDocument, ite)
	{
		Document* pDoc = ite->second;
		pDoc->p_StartingEnhancedState = pDoc->p_TrainingEnhancedState;
	}
}


//													
void Document::SwitchToTestStateGraph (void)
{
	ITERATE (IdToDocument_map_t, map_IdToDocument, ite)
	{
		Document* pDoc = ite->second;
		pDoc->p_StartingEnhancedState = pDoc->p_TestEnhancedState;
	}
}



//													
void Document::DumpRewards (int _iIteration)
{
	{
		String sFileName;
		sFileName << (config)"train_value_dump_file" << _iIteration << ".values";
		ofstream file (sFileName, ios_base::out);
		ITERATE (IdToDocument_map_t, map_IdToDocument, ite)
		{
			Document* pDoc = ite->second;
			if (NULL == pDoc->p_TrainingEnhancedState)
				continue;
			file << "[document] " << pDoc->s_Id << endl;
			pDoc->p_TrainingEnhancedState->DumpRewards (file, 0);
		}
	}

	{
		String sFileName;
		sFileName << (config)"test_value_dump_file" << _iIteration << ".values";
		ofstream file (sFileName, ios_base::out);
		ITERATE (IdToDocument_map_t, map_IdToDocument, ite)
		{
			Document* pDoc = ite->second;
			if (NULL == pDoc->p_TestEnhancedState)
				continue;
			file << "[document] " << pDoc->s_Id << endl;
			pDoc->p_TestEnhancedState->DumpRewards (file, 0);
		}
	}
}



//													
Document* Document::Find (const String& _sId)
{
	IdToDocument_map_t::iterator	ite;
	ite = map_IdToDocument.find (_sId);
	if (map_IdToDocument.end () == ite)
		return NULL;

	return ite->second;
}



//													
String_dq_t* Document::GetIdsForStartingState (String& _sStartingState)
{
	StartingStateToIds_map_t::iterator	ite;
	ite = map_StartingStateToIds.find (_sStartingState);
	if (map_StartingStateToIds.end () == ite)
		return NULL;

	return ite->second;
}


//													
std::ostream& operator<< (std::ostream& _rStream, const Document& _rDocument)
{
	Document& rDocument = (Document&) _rDocument;

	_rStream << "doc [id=" << rDocument.s_Id  << "] "
			 << "[start state=" << rDocument.s_RequiredStartingState << "] "
			 << "[sentences=" << rDocument.vec_Sentences.size () << "]" << endl;

	ITERATE (Sentence_vec_t, rDocument.vec_Sentences, ite)
		_rStream << **ite << endl;
	_rStream << endl;
	
	return _rStream;
}



//													
//													
TextInterface::TextInterface (void)
{
	p_Document = NULL;

	set_UsedWords.clear ();
	i_LastLocation_WU = -1;
	i_LastLocation_WC = -1;
	i_LastLocation_WP = -1;
}


//													
TextInterface::~TextInterface (void)
{
	assert ((Document*)0x1 != p_Document);
	p_Document = (Document*)0x1;
}


//													
bool TextInterface::StartEpisode (String& _rDocumentId)
{
	p_Document = Document::Find (_rDocumentId);
    if (NULL == p_Document)
        return false;
    if (true == p_Document->vec_Sentences.empty ())
        return false;

	set_UsedWords.clear ();
	i_LastLocation_WU = -1;
	i_LastLocation_WC = -1;
	i_LastLocation_WP = -1;

	return true;
}



//													
void TextInterface::EndEpisode (void)
{
	p_Document = NULL;

	set_UsedWords.clear ();
	i_LastLocation_WU = -1;
	i_LastLocation_WC = -1;
	i_LastLocation_WP = -1;
}



//													
void TextInterface::UseWords (char _lwc, char _lwu, char _lwp)
{
	set_UsedWords.insert (_lwu);
	set_UsedWords.insert (_lwc);
	 if (-1 != _lwp)
	 	set_UsedWords.insert (_lwp);

	i_LastLocation_WC = _lwc;
	i_LastLocation_WU = _lwu;
	i_LastLocation_WP = _lwp;
}



//													
void TextInterface::UseWordSpan (char _lwc, char _lwu, char _lwp)
{
	int iLower = (_lwc < _lwu)? _lwc : _lwu;
	int iHigher = (_lwc > _lwu)? _lwc : _lwu;
	if (-1 != _lwp)
	{
		iLower = (iLower < _lwp)? iLower : _lwp;
		iHigher = (iHigher > _lwp)? iHigher : _lwp;
	}

	for (int i = iLower; i <= iHigher; ++ i)
		set_UsedWords.insert (i);

	i_LastLocation_WC = _lwc;
	i_LastLocation_WU = _lwu;
	i_LastLocation_WP = _lwp;
}



