#include <assert.h>
#include <nlp_config.h>
#include <nlp_macros.h>
#include <nlp_time.h>
#include "FieldId.h"
#include "ExperienceInterface.h"
#include "State.h"


//											
ExperienceInterface::ExperienceInterface (void)
{
	b_ConnectedToEnvironment = false;
	b_InEpisode = false;
	i_ActionCount = 0;
	b_StateChangedSinceLastObservation = true;
	p_LastObservedState = NULL;
}



//											
ExperienceInterface::~ExperienceInterface (void)
{
	if (true == b_ConnectedToEnvironment)
	{
		cout << "[WARNING]  destructor called on ExperienceInterface "
				" while it is still connected to the environment.  Call "
				" ExperienceInterface::Stop () before deleting object."
			 << endl;
		ExperienceInterface::Stop ();
	}

	b_ConnectedToEnvironment = false;
	b_InEpisode = false;
	i_ActionCount = -1;
	p_LastObservedState = (GraphCacheNode*)0x1;
}



//											
bool ExperienceInterface::IsConnectedToEnvironment (void)
{
	return b_ConnectedToEnvironment;
}



//											
bool ExperienceInterface::Start (void)
{
	o_Cache.Load ();
	if (false == o_EnvironmentInterface.Start ())
		return false;
	CacheInterface::Init (&o_Cache, &o_EnvironmentInterface);
	b_ConnectedToEnvironment = true;
	return true;
}



//											
void ExperienceInterface::Stop (void)
{
	if (true == b_InEpisode)
	{
		cout << "[WARNING]  ExperienceInterface::Stop () "
				"called in the middle of an episode."
			 << endl;
		EndEpisode ();
	}
	b_ConnectedToEnvironment = false;
	
	o_Cache.Save ();
	o_EnvironmentInterface.Stop ();
}



//											
bool ExperienceInterface::KnownWorldOnly (void)
{
	String sRequest ("[known world only]\n");
	String sResponse;
	if (false == o_EnvironmentInterface.SendReceive (sRequest, &sResponse))
		return false;

	if (false == sResponse.StartsWith ("Known-world-only mode active"))
	{
		cout << "[ERROR]  Failed to switch to known-world-only mode: "
			 << sResponse << endl;
		return false;
	}
	return true;
}



//											
bool ExperienceInterface::StartEpisode (const char* _zRequiredStartingState)
{
	if (false == b_ConnectedToEnvironment)
	{
		cout << "EnvironmentState::StartEpisode () called, but the "
				"environment interface has not yet been initiated."
			 << endl;
		return false;
	}

	b_InEpisode = false;

	String sStartState (_zRequiredStartingState);
	CacheInterface::StartEpisode (sStartState);

	b_StateChangedSinceLastObservation = true;
	p_LastObservedState = NULL;

	b_InEpisode = true;
	i_ActionCount = 0;
	return true;
}



//											
bool ExperienceInterface::EndEpisode (void)
{
	if (false == b_InEpisode)
	{
		cout << "[ERROR]  ExperienceInterface::EndEpisode () called before StartEpisode ()."
			 << endl;
		return false;
	}

	// The cache interfaces saves the current trace		
	// to the cache when EndEpisode is called. At this	
	// point, it might detect a cache-live mismatch, 	
	// and discard the nodes associated with the trace.	
	// It returns false in this case, and we should		
	// not process this trace (compute reward, etc) and	
	// further.  Returning false from this method lets	
	// the caller know of this.							
	bool bEndEpisodeSuccessful = CacheInterface::EndEpisode ();

	// update cache node pointers in the enhanced graph
	ITERATE (CacheNode_dq_t, dq_TemporaryCacheNodes, ite)
	{
		GraphCacheNode* pOldNode = (GraphCacheNode*) *ite;
		GraphCacheNode* pNewNode 
			= (GraphCacheNode*) o_Cache.FindNodeForState (pOldNode->GetState ());

		EnhancedState::UpdateAllCacheNodes (pOldNode, pNewNode);
	}

	b_InEpisode = false;
	return bEndEpisodeSuccessful;
}


//											
void ExperienceInterface::ClearEpisodeData (void)
{
	ITERATE (CacheNode_dq_t, dq_TemporaryCacheNodes, ite)
		delete *ite;
	dq_TemporaryCacheNodes.clear ();
}


//											
GraphCacheNode* ExperienceInterface::MeasureState (void)
{
	if (false == b_InEpisode)
	{
		cout << "[ERROR]  ExperienceInterface::MeasureState () called before StartEpisode ()."
			 << endl;
		return NULL;
	}

	GraphCacheNode* pMeasuredState;
	if (true == b_StateChangedSinceLastObservation)
	{
		String sCommand ("*\n");
		pMeasuredState = (GraphCacheNode*) CacheInterface::SendReceive (sCommand, &dq_TemporaryCacheNodes);
		p_LastObservedState = pMeasuredState;
	}
	else
	{
		assert (NULL != p_LastObservedState);
		pMeasuredState = p_LastObservedState;
	}
	b_StateChangedSinceLastObservation = false;

	return pMeasuredState;
}



//											
EnvResponseToAction_e ExperienceInterface::SendAction (String& _rAction)
{
	if (false == b_InEpisode)
	{
		cout << "[ERROR]  ExperienceInterface::SendAction () called before StartEpisode ()."
			 << endl;
		return er_error;
	}

	b_StateChangedSinceLastObservation = true;
	p_LastObservedState = NULL;

	String sResponse;
	sResponse = CacheInterface::SendReceive (_rAction);
	if (true == sResponse.Has ("[ERROR]"))
	{
		cout << "[WARNING] Error response received from OS agent." << endl;
		cout << "          command ->" << _rAction << "<-" << endl;
		cout << "          response:" << endl;
		cout << "------------------------------------------------" << endl;
		cout << sResponse << endl;
		cout << "------------------------------------------------" << endl;
		
		return er_error;
	}
	
	++ i_ActionCount;
	if (true == sResponse.Has ("env has changed"))
		return er_env_changed;
	else
		return er_env_unchanged;
}


//											
void ExperienceInterface::LockCacheForReading (void)
{
	o_Cache.LockForReading ();
	c_CacheLockMode = CACHE_LOCK_READ;
}


//											
void ExperienceInterface::UnlockCache (void)
{
	c_CacheLockMode = CACHE_LOCK_NONE;
	o_Cache.Unlock ();
}


//											
GraphCacheNode* ExperienceInterface::PeekAction (String& _rAction, CacheNode* _pCurrentNode)
{
	CacheNode* pNextNode = NULL;
	if (NULL != _pCurrentNode)
		pNextNode = _pCurrentNode->FindChild (_rAction);
	else if (NULL != p_CurrentNode)
		pNextNode = p_CurrentNode->FindChild (_rAction);
	else
		pNextNode = NULL;

	return (GraphCacheNode*) pNextNode;
}


//											
bool ExperienceInterface::SaveExperience (void)
{
	return o_Cache.Save ();
}
