#include <nlp_macros.h>
#include <nlp_config.h>
#include "State.h"
#include "FieldId.h"
#include "Learner.h"


CacheNodeToEnhancedState_mmp_t	EnhancedState::mmp_CacheNodeToEnhancedState;
StringToEditDist_map_t			RLState::map_StringToNormalizedEditDist;
double							RLState::d_EditDistanceCutoff = -1;
double							RLState::d_SubStringEditDistanceCutoff = -1;


//																
EnvironmentState::EnvironmentState (void)
{
	p_EnvironmentState = NULL;
}


EnvironmentState::~EnvironmentState (void)
{
	assert ((GraphCacheNode*)0x1 != p_EnvironmentState);
	p_EnvironmentState = (GraphCacheNode*)0x1;
}


//																
EnvResponseToAction_e EnvironmentState::ApplyEnvAction (ExperienceInterface* _pInterface,
														EnvironmentAction& _rAction)
{
	String sCommandType = Command::Get (_rAction.i_SelectedCommandIndex);
	if ("[none]" == sCommandType)
		return er_env_unchanged;

	String sCommand;
	sCommand << '!' << _rAction.p_SelectedObject->z_ID << ':' 
			 << Command::Get (_rAction.i_SelectedCommandIndex);
	if (true == Command::NeedsParam (_rAction.i_SelectedCommandIndex))
	{
		if ("" == _rAction.s_SelectedParameter)
		{
			cout << "[ERROR]  The command parameter is empty for a command ("
				 << Command::Get (_rAction.i_SelectedCommandIndex) << ") that needs a parameter."
				 << endl;
		}
		sCommand << '|' << _rAction.s_SelectedParameter;
	}
	sCommand << '\n';

	return _pInterface->SendAction (sCommand);
}


//																
EnvResponse_e EnvironmentState::ObserveEnvState (ExperienceInterface* _pInterface)
{
	GraphCacheNode* pEnvironmentState = _pInterface->MeasureState ();
	if (NULL == pEnvironmentState)
		return er_env_error;

	bool bHasVisibleObjects = false;
	int iStateObjectCount = pEnvironmentState->GetObjectCount ();
	for (int i = 0; i < iStateObjectCount; ++ i)
	{
		UIObject* pObject = (UIObject*) pEnvironmentState->GetObject (i);
		if (1 == pObject->bit_Features [VISIBLE])
			bHasVisibleObjects = true;
	}


	if (NULL != p_EnvironmentState)
	{
		if (pEnvironmentState != p_EnvironmentState)
			cout << "[WARNING]  Observed state does not match with memory." << endl;

		if (pEnvironmentState != p_EnvironmentState)
			UpdateOwnCacheNode (pEnvironmentState);
	}
	else
	{
		p_EnvironmentState = pEnvironmentState;
		RememberCacheNode ();
	}

	return (true == bHasVisibleObjects)? er_ok : er_no_visible_objects;
}


//																
std::ostream& operator<< (std::ostream& _rStream, const EnvironmentState& _rEnv)
{
	for (int i = 0; i < ((EnvironmentState&)_rEnv).p_EnvironmentState->GetObjectCount (); ++ i)
	{
		UIObject* pObject = (UIObject*) ((EnvironmentState&)_rEnv).p_EnvironmentState->GetObject (i);
		_rStream << *pObject << endl;
	}
	return _rStream;
}




//																
//																
TextState::TextState (void)
{
	p_Document = NULL;
	p_UnusedWordLocations = NULL;
	i_UnusedWordLocationCount = 0;
}


TextState::~TextState (void)
{
	if (NULL != p_UnusedWordLocations)
		delete[] p_UnusedWordLocations;
	assert (-1 != i_UnusedWordLocationCount);
	i_UnusedWordLocationCount = -1;
	set_PreviouslyUsedWords.clear ();
}


//																
bool TextState::WordLocationsAreValid (char _lWC, char _lWU, char _lWP)
{
	int iSentenceWC = p_Document->GetSentenceIndex (_lWC);
	int iSentenceWU = p_Document->GetSentenceIndex (_lWU);
	if (iSentenceWC != iSentenceWU)
		return false;
	if (_lWP != p_Document->i_WordCount)
	{
		int iSentenceWP = p_Document->GetSentenceIndex (_lWP);
		if (iSentenceWC != iSentenceWP)
			return false;
	}

	char iLeft = (_lWC < _lWU)? _lWC : _lWU;
	char iRight = (_lWC > _lWU)? _lWC : _lWU;
	if (p_Document->i_WordCount != _lWP)
	{
		iLeft = (iLeft < _lWP)? iLeft : _lWP;
		iRight = (iRight > _lWP)? iRight : _lWP;
	}

	char_set_t::iterator	iteUpperBound;
	iteUpperBound = set_PreviouslyUsedWords.upper_bound (iLeft);
	if (set_PreviouslyUsedWords.end () == iteUpperBound)
		return true;

	if (*iteUpperBound <= iRight)
		return false;

	return true;
}


//																
void TextState::ObserveTextState (TextInterface* _pInterface)
{
	Document* pDocument = _pInterface->GetCurrentDocument ();

	char iPreviousLocation_WC = _pInterface->LastUsedCommandWord ();
	char iPreviousLocation_WU = _pInterface->LastUsedObjectWord ();
	char iPreviousLocation_WP = _pInterface->LastUsedParameterWord ();

	_pInterface->GetUsedWords (&set_PreviouslyUsedWords);

	char iUnusedWordLocationCount = pDocument->i_WordCount
									- set_PreviouslyUsedWords.size ();
	char* pUnusedWordLocations = new char [iUnusedWordLocationCount];
	int iIndex = 0;
	for (char i = 0; i < pDocument->i_WordCount; ++ i)
	{
		if (set_PreviouslyUsedWords.end () == set_PreviouslyUsedWords.find (i))
			pUnusedWordLocations [iIndex ++] = i;
	}


	if (NULL != p_Document)
	{
		assert (pDocument == p_Document);
		assert (iPreviousLocation_WC == i_PreviousLocation_WC);
		assert (iPreviousLocation_WU == i_PreviousLocation_WU);
		assert (iPreviousLocation_WP == i_PreviousLocation_WP);
		assert (iUnusedWordLocationCount == i_UnusedWordLocationCount);
		assert (0 == memcmp (pUnusedWordLocations, 
							 p_UnusedWordLocations,
							 iUnusedWordLocationCount * sizeof (char)));

		delete[] pUnusedWordLocations;
	}
	else
	{
		p_Document = pDocument;
		i_PreviousLocation_WC = iPreviousLocation_WC;
		i_PreviousLocation_WU = iPreviousLocation_WU;
		i_PreviousLocation_WP = iPreviousLocation_WP;
		i_UnusedWordLocationCount = iUnusedWordLocationCount;
		p_UnusedWordLocations = pUnusedWordLocations;
	}
}


//																
void TextState::ApplyTextAction (TextInterface* _pInterface,
								 TextAction& _rAction,
								 bool _bTrainingMode)
{
	if (_rAction.i_SelectedObjectWordLocation == p_Document->i_WordCount)
		return;

	dq_PreviouslyUsedWU.push_back (_rAction.i_SelectedObjectWordLocation);

	if (_rAction.i_SelectedParameterWordLocation != p_Document->i_WordCount)
	{
		if (true == _bTrainingMode)
			_pInterface->UseWordSpan (_rAction.i_SelectedCommandWordLocation,
									  _rAction.i_SelectedObjectWordLocation,
									  _rAction.i_SelectedParameterWordLocation);
		else
			_pInterface->UseWords (_rAction.i_SelectedCommandWordLocation,
									  _rAction.i_SelectedObjectWordLocation,
									  _rAction.i_SelectedParameterWordLocation);
	}
	else
	{
		if (true == _bTrainingMode)
			_pInterface->UseWordSpan (_rAction.i_SelectedCommandWordLocation,
									  _rAction.i_SelectedObjectWordLocation,
									  -1);
		else
			_pInterface->UseWords (_rAction.i_SelectedCommandWordLocation,
									  _rAction.i_SelectedObjectWordLocation,
									  -1);
	}
}




//																
//																
EnhancedState::EnhancedState (void)
{
	p_Parent = NULL;
	d_Reward = NAN;
}


//																
EnhancedState::~EnhancedState (void)
{
	assert ((EnhancedState*)0x1 != p_Parent);
	
	if (NULL != p_Parent)
		p_Parent->DiscardChild (this);

	// dissociate from cache->state map...	
	pair <CacheNodeToEnhancedState_mmp_t::iterator,
		  CacheNodeToEnhancedState_mmp_t::iterator> pairResult;
	pairResult = mmp_CacheNodeToEnhancedState.equal_range (p_EnvironmentState);

	CacheNodeToEnhancedState_mmp_t::iterator	ite;
	for (ite = pairResult.first; ite != pairResult.second;)
	{
		EnhancedState* pState = ite->second;
		if (pState != this)
		{
			++ ite;
			continue;
		}

		// post increment below is critical - we want the	
		// iterator to be incremented before erase is called
		// since the iterator is not valid afterwards...	
		mmp_CacheNodeToEnhancedState.erase (ite ++);
	}

	// we need to do the stuff below because	
	// the destructor (this method) will call	
	// Discard child...							
	set<void*> setStatesToDelete;
	ITERATE (ActionToEnhancedState_hmp_t, hmp_ActionToEnhancedState, ite)
		setStatesToDelete.insert (ite->second);
	hmp_ActionToEnhancedState.clear ();

	ITERATE (set<void*>, setStatesToDelete, ite)
		delete (EnhancedState*) *ite;


	//											
	d_Reward = NAN;
	p_Parent = (EnhancedState*)0x1;
}


//																
void EnhancedState::AddTransition (long _lAction, EnhancedState* _pNextState)
{
	assert (hmp_ActionToEnhancedState.end () 
			== hmp_ActionToEnhancedState.find (_lAction));

	_pNextState->p_Parent = this;
	hmp_ActionToEnhancedState.insert (make_pair (_lAction, _pNextState));
}


//																
void EnhancedState::DiscardChild (EnhancedState* _pChild)
{
	_pChild->p_Parent = NULL;

	// On deletion, the state's children are also	
	// deleted, which forces a call to here, even	
	// though hmp_ActionToEnhancedState would be	
	// empty.  Check and return in this case...		
	if (true == hmp_ActionToEnhancedState.empty ())
		return;

	deque <long> dqActions;
	ITERATE (ActionToEnhancedState_hmp_t, hmp_ActionToEnhancedState, ite)
	{
		EnhancedState* pNextState = ite->second;
		if (pNextState != _pChild)
			continue;

		dqActions.push_back (ite->first);
	}

	assert (false == dqActions.empty ());
	ITERATE (deque<long>, dqActions, iteAction)
		hmp_ActionToEnhancedState.erase (*iteAction);
}


//																
void EnhancedState::DumpRewards (ostream& _rStream, int _iDepth)
{
	for (int i = 0; i < _iDepth; ++i)
		_rStream << ',';
	_rStream << d_Reward << '\n';

	ITERATE (ActionToEnhancedState_hmp_t, hmp_ActionToEnhancedState, ite)
	{
		EnhancedState* pChild = ite->second;
		pChild->DumpRewards (_rStream, _iDepth + 1);
	}
}


//																
EnhancedState* EnhancedState::GetNextState (long _lAction)
{
	ActionToEnhancedState_hmp_t::iterator	ite;
	ite = hmp_ActionToEnhancedState.find (_lAction);
	if (hmp_ActionToEnhancedState.end () == ite)
		return NULL;
	
	return ite->second;
}


//																
void EnhancedState::RememberCacheNode (void)
{
	mmp_CacheNodeToEnhancedState.insert (make_pair (p_EnvironmentState, this));
}


//																
void EnhancedState::DiscardCacheNode (GraphCacheNode* _pNode)
{
	pair <CacheNodeToEnhancedState_mmp_t::iterator,
		  CacheNodeToEnhancedState_mmp_t::iterator> pairResult;
	pairResult = mmp_CacheNodeToEnhancedState.equal_range (_pNode);
	if (pairResult.first == pairResult.second)
		return;

	set<void*> setStatesToDelete;
	CacheNodeToEnhancedState_mmp_t::iterator	ite;
	for (ite = pairResult.first; ite != pairResult.second; ++ ite)
	{
		EnhancedState* pState = ite->second;
		if (NULL != pState->p_Parent)
			pState->p_Parent->DiscardChild (pState);
		setStatesToDelete.insert (pState);
	}

	mmp_CacheNodeToEnhancedState.erase (pairResult.first, pairResult.second);

	ITERATE (set<void*>, setStatesToDelete, ite)
		delete (EnhancedState*) *ite;
}


//																
void EnhancedState::UpdateOwnCacheNode (GraphCacheNode* _pNewNode)
{
	pair <CacheNodeToEnhancedState_mmp_t::iterator,
		  CacheNodeToEnhancedState_mmp_t::iterator> pairResult;
	pairResult = mmp_CacheNodeToEnhancedState.equal_range (p_EnvironmentState);
	if (pairResult.first == pairResult.second)
	{
		p_EnvironmentState = _pNewNode;
		RememberCacheNode ();
		return;
	}

	if (NULL != _pNewNode)
	{
		CacheNodeToEnhancedState_mmp_t::iterator	ite;
		for (ite = pairResult.first; ite != pairResult.second; ++ ite)
		{
			EnhancedState* pState = ite->second;
			if (this != pState)
				continue;

			mmp_CacheNodeToEnhancedState.erase (ite);

			p_EnvironmentState = _pNewNode;
			RememberCacheNode ();
			break;
		}
	}
	else
	{
		// this should never happen !  It deletes *this*,	
		// which should cause havoc in the caller...		
		abort ();

		CacheNodeToEnhancedState_mmp_t::iterator	ite;
		for (ite = pairResult.first; ite != pairResult.second; ++ ite)
		{
			EnhancedState* pState = ite->second;
			if (this != pState)
				continue;

			mmp_CacheNodeToEnhancedState.erase (ite);

			if (NULL != pState->p_Parent)
				pState->p_Parent->DiscardChild (pState);
			delete pState;
			break;
		}
	}
}


//																
void EnhancedState::UpdateAllCacheNodes (GraphCacheNode* _pOldNode, 
										 GraphCacheNode* _pNewNode)
{
	if (_pOldNode == _pNewNode)
		return;

	pair <CacheNodeToEnhancedState_mmp_t::iterator,
		  CacheNodeToEnhancedState_mmp_t::iterator> pairResult;
	pairResult = mmp_CacheNodeToEnhancedState.equal_range (_pOldNode);
	if (pairResult.first == pairResult.second)
		return;


	if (NULL != _pNewNode)
	{
		EnhancedState_dq_t	dqUpdatedStates;

		// update env state pointers...				
		CacheNodeToEnhancedState_mmp_t::iterator	ite;
		for (ite = pairResult.first; ite != pairResult.second; ++ ite)
		{
			EnhancedState* pState = ite->second;
			assert (pState->p_EnvironmentState->GetNormalizedState ()
					== _pNewNode->GetNormalizedState ());

			pState->p_EnvironmentState = _pNewNode;
			dqUpdatedStates.push_back (pState);
		}

		// forget old env state -> state mapping...	
		mmp_CacheNodeToEnhancedState.erase (pairResult.first, pairResult.second);


		// remember states using new env state...	
		ITERATE (EnhancedState_dq_t, dqUpdatedStates, iteState)
		{
			EnhancedState* pState = *iteState;
			pState->RememberCacheNode ();
		}
	}
	else
	{
		set<void*> setStatesToDelete;
		CacheNodeToEnhancedState_mmp_t::iterator	ite;
		for (ite = pairResult.first; ite != pairResult.second; ++ ite)
		{
			EnhancedState* pState = ite->second;
			if (NULL != pState->p_Parent)
				pState->p_Parent->DiscardChild (pState);
			setStatesToDelete.insert (pState);
		}
		mmp_CacheNodeToEnhancedState.erase (pairResult.first, pairResult.second);

		ITERATE (set<void*>, setStatesToDelete, ite)
			delete (EnhancedState*) *ite;
	}

}


//																
void EnhancedState::DiscardBranch (GraphCacheNode* _pFrom, GraphCacheNode* _pTo)
{
	pair <CacheNodeToEnhancedState_mmp_t::iterator,
		  CacheNodeToEnhancedState_mmp_t::iterator> pairResult;
	pairResult = mmp_CacheNodeToEnhancedState.equal_range (_pFrom);
	if (pairResult.first == pairResult.second)
		return;


	CacheNodeToEnhancedState_mmp_t::iterator	ite;
	for (ite = pairResult.first; ite != pairResult.second; ++ ite)
	{
		EnhancedState* pState = ite->second;

		deque <long> dqActions;
		deque <EnhancedState*> dqStates;
		ITERATE (ActionToEnhancedState_hmp_t, pState->hmp_ActionToEnhancedState, iteState)
		{
			EnhancedState* pNextState = iteState->second;
			if (pNextState->p_EnvironmentState != _pTo)
				continue;

			dqActions.push_back (iteState->first);
			dqStates.push_back (pNextState);
		}

		ITERATE (deque<EnhancedState*>, dqStates, iteState)
			delete *iteState;

		ITERATE (deque<long>, dqActions, iteAction)
			pState->hmp_ActionToEnhancedState.erase (*iteAction);
	}
}



//																
void EnhancedState::DebugPrint (void)
{
	cout << d_Reward << '\n';

	ITERATE (ActionToEnhancedState_hmp_t, hmp_ActionToEnhancedState, ite)
	{
		EnhancedState* pChild = ite->second;
		cout << pChild->d_Reward;

		RLAction oAction = RLAction::LongToAction (ite->first);
		String sWC (p_Document->sGetWord (oAction.i_SelectedCommandWordLocation));
		String sWU (p_Document->sGetWord (oAction.i_SelectedObjectWordLocation));
		int iWP = -1;
		String sWP ("(none)");
		if (oAction.i_SelectedParameterWordLocation < p_Document->i_WordCount)
		{
			iWP = oAction.i_SelectedParameterWordLocation;
			sWP = p_Document->sGetWord (oAction.i_SelectedParameterWordLocation);
		}
		
		UIObject* pObject = (UIObject*) p_EnvironmentState->GetObject (oAction.i_SelectedObjectIndex);

		cout << "  - wc: " << oAction.i_SelectedCommandWordLocation << '|' << sWC 
			 << ", wu: " << oAction.i_SelectedObjectWordLocation << '|' << sWU 
			 << ", wp: " << iWP << '|' << sWP;
		cout << ",  c: " << Command::Get (oAction.i_SelectedCommandIndex);
		cout << ",  u: " << pObject->s_CleanName << endl;
	}
}



//																
//																
RLState::RLState (void)
{
	p_EnhancedState = NULL;
	p_UnusedPotentialObjectWords = NULL;
	i_UnusedPotentialObjectWordCount = 0;
}

RLState::~RLState (void)
{
	hmp_WordsInEnvironment.clear ();
	hmp_WordToLowestEditDistance.clear ();
	vec_WordObjects.clear ();

	if (NULL != p_UnusedPotentialObjectWords)
		delete[] p_UnusedPotentialObjectWords;
	assert (-1 != i_UnusedPotentialObjectWordCount);
	i_UnusedPotentialObjectWordCount = -1;
}


//																
void RLState::Init (void)
{
	d_EditDistanceCutoff = (config)"edit_distance_cutoff";
	d_SubStringEditDistanceCutoff = (config)"substring_edit_distance_cutoff";
}


//																
EnvResponse_e RLState::ObserveState (EnhancedState* _pCurrentState)
{
	p_EnhancedState = _pCurrentState;
	assert (NULL != p_EnhancedState);

	// ObserveEnvState () and ObserveTextState () do	
	// not need to be called since all that information	
	// is already present in p_EnhancedState...			

	IdentifyUnusedPotentialObjectWords ();
	return (i_UnusedPotentialObjectWordCount > 0)? er_ok : er_no_object_words;
}


//																
EnvResponse_e RLState::ObserveState (EnhancedState* _pCurrentState, 
							ExperienceInterface* _pExperienceInterface,
							TextInterface* _pTextInterface)
{
	p_EnhancedState = _pCurrentState;
	assert (NULL != p_EnhancedState);

	EnvResponse_e eResponse = p_EnhancedState->ObserveEnvState (_pExperienceInterface);
	if (er_ok != eResponse)
		return eResponse;
	
	p_EnhancedState->ObserveTextState (_pTextInterface);
	IdentifyUnusedPotentialObjectWords ();
	return (i_UnusedPotentialObjectWordCount > 0)? er_ok : er_no_object_words;
}


//																
EnvResponseToAction_e RLState::ApplyAction (ExperienceInterface* _pExperienceInterface,
											TextInterface* _pTextInterface,
											RLAction& _rAction,
											bool _bTrainingMode)
{
	EnvResponseToAction_e eResponse 
		= p_EnhancedState->ApplyEnvAction (_pExperienceInterface, _rAction);
	if (er_error == eResponse)
	{
		Learner::ofs_Log << "[ERROR]  ApplyEnvAction failed." << endl;
		return er_error;
	}
	p_EnhancedState->ApplyTextAction (_pTextInterface, _rAction, _bTrainingMode);
	return eResponse;
}


//																
String RLState::ConvertToCommandString (RLAction& _rAction)
{
	_rAction.p_SelectedObject = GetObject (_rAction.i_SelectedObjectIndex);
	if (_rAction.i_SelectedParameterWordLocation < TextLength ())
		_rAction.s_SelectedParameter = sGetWord (_rAction.i_SelectedParameterWordLocation);

	String sCommand;
	sCommand << '!' << _rAction.p_SelectedObject->z_ID << ':' << Command::Get (_rAction.i_SelectedCommandIndex);
	if (true == Command::NeedsParam (_rAction.i_SelectedCommandIndex))
	{
		if ("" == _rAction.s_SelectedParameter)
		{
			cout << "[ERROR]  The command parameter is empty for a command ("
				 << Command::Get (_rAction.i_SelectedCommandIndex) << ") that needs a parameter."
				 << endl;
		}
		sCommand << '|' << _rAction.s_SelectedParameter;
	}
	sCommand << '\n';
	return sCommand;
}


//																
void RLState::IdentifyUnusedPotentialObjectWords (void)
{
	assert (true == vec_WordObjects.empty ());
	assert (true == hmp_WordsInEnvironment.empty ());
	assert (NULL == p_UnusedPotentialObjectWords);

	vec_WordObjects.resize (TextLength ());

	hmp_WordsInEnvironment.clear ();

	// unused potential words is going to be atleast as big as the unused words ...
	p_UnusedPotentialObjectWords = new int [UnusedWordCount ()];
	i_UnusedPotentialObjectWordCount = 0;
	
	for (int i = 0; i < UnusedWordCount (); ++ i)
	{
		char lw = UnusedWordLocation (i);
		String sWord = sGetWord (lw);

		UIObject_dq_t& rdqObjects = vec_WordObjects [lw];
		for (int u = 0; u < p_EnhancedState->p_EnvironmentState->GetObjectCount (); ++ u)
		{
			UIObject* pObject = (UIObject*) p_EnhancedState->p_EnvironmentState->GetObject (u);
			if (0 == pObject->bit_Features [VISIBLE])
				continue;
			if ("" == pObject->s_CleanName)
				continue;
			if (true == RLState::WordMaybeObject (sWord, pObject->s_CleanName))
				rdqObjects.push_back (pObject);
		}

		if (false == rdqObjects.empty ())
		{
			p_UnusedPotentialObjectWords [i_UnusedPotentialObjectWordCount ++] = lw;
			hmp_WordsInEnvironment.insert (make_pair (sWord, true));
		}
		else
			hmp_WordsInEnvironment.insert (make_pair (sWord, false));
	}
}


//																
void RLState::IdentifyLowestEditDistances (void)
{
	for (int i = 0; i < UnusedWordCount (); ++ i)
	{
		String sWord = sGetWord (UnusedWordLocation (i));
		// the LowestEditDistanceInState method call	
		// computes & caches the edit distances...		
		LowestEditDistanceInState (sWord);
	}
}


//																
bool RLState::CurrentStateHasWord (String& _sWord)
{
	WordsInEnvironment_hmp_t::iterator	ite;
	ite = hmp_WordsInEnvironment.find (_sWord);
	if (hmp_WordsInEnvironment.end () != ite)
		return ite->second;

	// we should never get here because IdentifyUnusedPotentialObjectWords	
	// should have put everything in the cache by now...					
	assert (false);

	for (int i = 0; i < p_EnhancedState->p_EnvironmentState->GetObjectCount (); ++ i)
	{
		UIObject* pObject = (UIObject*) p_EnhancedState->p_EnvironmentState->GetObject (i);
		if (0 == pObject->bit_Features [VISIBLE])
			continue;
		if ("" == pObject->s_CleanName)
			continue;
		if (true == WordMaybeObject (_sWord, pObject->s_CleanName))
		{
			hmp_WordsInEnvironment.insert (make_pair (_sWord, true));
			return true;
		}
	}

	hmp_WordsInEnvironment.insert (make_pair (_sWord, false));
	return false;
}


//																
bool RLState::WordMaybeObject (String _sWord, String _sObjectName)
{
	_sWord.ReplacePattern ("[^a-zA-Z0-9 ]+", "");
	_sWord.LowerCase ();
	_sObjectName.ReplacePattern ("[^a-zA-Z0-9 ]+", "");
	_sObjectName.LowerCase ();

	if (("" == _sWord) || ("" == _sObjectName))
		return false;
	if (_sObjectName == _sWord)
		return true;

	double dDistance = RLState::NormalizedEditDistance (_sObjectName, _sWord);

	// if the object has low edit distance, 		
	// we don't care about substrings				
	if (dDistance < d_EditDistanceCutoff)
		return true;

	// if the object has high edit distance, 		
	// we want to check substrings					
	if (dDistance < d_SubStringEditDistanceCutoff)
	{
		if (_sObjectName == _sWord)
			return true;
		if (true == _sObjectName.StartsWith (_sWord + ' '))
			return true;
		if (true == _sObjectName.EndsWith (' ' + _sWord))
			return true;
		if (true == _sObjectName.Has (' ' + _sWord + ' '))
			return true;
	}

	return false;
}


//																
bool RLState::WordMaybeObjectStrict (String _sWord, String _sObjectName)
{
	_sWord.ReplacePattern ("[^a-zA-Z0-9 ]+", "");
	_sWord.LowerCase ();
	_sObjectName.ReplacePattern ("[^a-zA-Z0-9 ]+", "");
	_sObjectName.LowerCase ();

	return (_sWord == _sObjectName);
}


//																
double RLState::LowestEditDistanceInState (String& _sWord)
{
	WordToLowestEditDistance_hmp_t::iterator	ite;
	ite = hmp_WordToLowestEditDistance.find (_sWord);
	if (hmp_WordToLowestEditDistance.end () != ite)
		return ite->second;

	double dLowestEditDistance = 1000;
	for (int i = 0; i < p_EnhancedState->p_EnvironmentState->GetObjectCount (); ++ i)
	{
		UIObject* pObject = (UIObject*) p_EnhancedState->p_EnvironmentState->GetObject (i);
		double dDistance = RLState::NormalizedEditDistance (pObject->s_CleanName, _sWord);
		if (dLowestEditDistance > dDistance)
			dLowestEditDistance = dDistance;
	}

	hmp_WordToLowestEditDistance.insert (make_pair (_sWord, dLowestEditDistance));
	return dLowestEditDistance;
}


//																
double RLState::NormalizedEditDistance (String _sLeft, String _sRight)
{
	StringToEditDist_map_t::iterator	ite;
	String sKey1;
	String sKey2;
	{
		sKey1 << _sLeft << '\xff' << _sRight;
		ite = map_StringToNormalizedEditDist.find (sKey1);
		if (map_StringToNormalizedEditDist.end () != ite)
			return ite->second;
	}
	{
		sKey2 << _sRight << '\xff' << _sLeft;
		ite = map_StringToNormalizedEditDist.find (sKey2);
		if (map_StringToNormalizedEditDist.end () != ite)
			return ite->second;
	}

	_sLeft.ReplacePattern ("[^a-zA-Z0-9 ]+", "");
	_sLeft.LowerCase ();
	_sRight.ReplacePattern ("[^a-zA-Z0-9 ]+", "");
	_sRight.LowerCase ();

	size_t iLeft = _sLeft.size ();
	size_t iRight = _sRight.size ();
	size_t iMaxLen = (iLeft > iRight)? iLeft : iRight;

	double dNormalizedEditDist = String::LevensteinDistance (_sLeft, _sRight) 
			/ (double) iMaxLen;
	
	map_StringToNormalizedEditDist.insert (make_pair (sKey1, dNormalizedEditDist));
	map_StringToNormalizedEditDist.insert (make_pair (sKey2, dNormalizedEditDist));

	return dNormalizedEditDist;
}


//																
int_set_t RLState::PotentialObjectWordsCoveredByState (char* _pWordLocations, int _iWordCount)
{
	GraphCache* pCache = (GraphCache*) p_EnhancedState->p_EnvironmentState->GetCache ();

	int_set_t setObjectWords;
	if (NULL == p_EnhancedState->p_Document)
		return setObjectWords;

	for (int i = 0; i < _iWordCount; ++ i)
	{
		int lw = _pWordLocations [i];
		int w = iGetWord (lw);
		int_set_t* psetObjects = pCache->GetObjectsForWord (w);
		if (NULL == psetObjects)
			continue;
		if (true == psetObjects->empty ())
			continue;

		UIObject_dq_t& rdqObjects = vec_WordObjects [lw];
		if (false == rdqObjects.empty ())
			setObjectWords.insert (w);
	}

	return setObjectWords;
}


//																
int_set_t RLState::PotentialObjectWordsCoveredByStateStrict (char* _pWordLocations, int _iWordCount)
{
	GraphCache* pCache = (GraphCache*) p_EnhancedState->p_EnvironmentState->GetCache ();

	int_set_t setObjectWords;
	if (NULL == p_EnhancedState->p_Document)
		return setObjectWords;

	for (int i = 0; i < _iWordCount; ++ i)
	{
		int lw = _pWordLocations [i];
		int w = iGetWord (lw);
		String sWord = sGetWord (lw);

		int_set_t* psetObjects = pCache->GetObjectsForWordStrict (w);
		if (NULL == psetObjects)
			continue;
		if (true == psetObjects->empty ())
			continue;

		UIObject_dq_t& rdqObjects = vec_WordObjects [lw];
		ITERATE (UIObject_dq_t, rdqObjects, iteObject)
		{
			UIObject* pObject = *iteObject;
			if (true == RLState::WordMaybeObjectStrict (sWord, pObject->s_CleanName))
			{
				setObjectWords.insert (w);
				break;
			}
		}
	}

	return setObjectWords;
}





