#ifndef __RLL_STATE__
#define __RLL_STATE__

#include <set>
#include <map>
#include <hash_map>
#include <vector>
#include <nlp_filesystem.h>
#include "Document.h"
#include "Action.h"
#include "UIObject.h"
#include "ExperienceInterface.h"
using namespace std;


class EnhancedState;
typedef set <int>							int_set_t;
typedef set <char>							char_set_t;
typedef deque <char>						char_dq_t;
typedef deque <EnhancedState*>				EnhancedState_dq_t;
typedef hash_map <String, bool>				WordsInEnvironment_hmp_t;
typedef hash_map <String, double>			WordToLowestEditDistance_hmp_t;
typedef vector <UIObject_dq_t>				WordObjects_vec_t;
typedef hash_map <String, double>			StringToEditDist_map_t;
typedef hash_map <long, EnhancedState*>		ActionToEnhancedState_hmp_t;
typedef multimap <void*, EnhancedState*>	CacheNodeToEnhancedState_mmp_t;
typedef set <void*>							EnhancedState_set_t;



// --------------------------------------------------------------------------
enum EnvResponse_e
{
	er_ok,
	er_no_object_words,
	er_no_visible_objects,
	er_env_error
};


// --------------------------------------------------------------------------
class EnvironmentState
{
	public:
		GraphCacheNode*					p_EnvironmentState;

		EnvironmentState (void);
		virtual ~EnvironmentState (void);

		virtual void RememberCacheNode (void) = 0;
		virtual void UpdateOwnCacheNode (GraphCacheNode* _pNew) = 0;

		EnvResponse_e ObserveEnvState (ExperienceInterface* _pInterface);
		EnvResponseToAction_e ApplyEnvAction (ExperienceInterface* _pInterface,
											  EnvironmentAction& _rAction);
};

std::ostream& operator<< (std::ostream& _rStream, const EnvironmentState& _rEnv);



// --------------------------------------------------------------------------
class TextState
{
	public:
		char_dq_t					dq_PreviouslyUsedWU;
		char_set_t					set_PreviouslyUsedWords;
		Document*					p_Document;
		char*						p_UnusedWordLocations;
		char						i_PreviousLocation_WU;
		char						i_PreviousLocation_WC;
		char						i_PreviousLocation_WP;
		char						i_UnusedWordLocationCount;
		

		TextState (void);
		~TextState (void);

		bool WordLocationsAreValid (char _lWC, char _lWU, char _lWP);
	
		void ObserveTextState (void);
		void ObserveTextState (TextInterface* _pInterface);
		void ApplyTextAction (TextInterface* _pInterface,
							  TextAction& _rAction,
							  bool _bTrainingMode);
};



// --------------------------------------------------------------------------
class EnhancedState : public EnvironmentState, public TextState
{
	friend class RLState; 

	private:
		static CacheNodeToEnhancedState_mmp_t	mmp_CacheNodeToEnhancedState;
		EnhancedState*							p_Parent;
		float									d_Reward;


	public:
		ActionToEnhancedState_hmp_t		hmp_ActionToEnhancedState;

		EnhancedState (void);
		~EnhancedState (void);

		void AddTransition (long _lAction, EnhancedState* _pNextState);
		void DiscardChild (EnhancedState* _pChild);
		void DumpRewards (ostream& _rStream, int _iDepth);

		EnhancedState* GetNextState (long _lAction);
		double GetReward (void)
			{ return d_Reward; };
		void GetReachableNodes (GraphCacheNode_set_t& _rsetNodes)
			{ p_EnvironmentState->GetReachableNodes (_rsetNodes); };

		void RememberCacheNode (void);
		void UpdateOwnCacheNode (GraphCacheNode* _pNew);

		static void UpdateAllCacheNodes (GraphCacheNode* _pOldNode, 
									 	 GraphCacheNode* _pNewNode);
		static void DiscardBranch (GraphCacheNode* _pFrom,
								   GraphCacheNode* _pTo);
		static void DiscardCacheNode (GraphCacheNode* _pNode);

		void DebugPrint (void);
};



// --------------------------------------------------------------------------
class RLState
{
	private:
		static double					d_EditDistanceCutoff;
		static double					d_SubStringEditDistanceCutoff;
		static StringToEditDist_map_t	map_StringToNormalizedEditDist;
		WordsInEnvironment_hmp_t		hmp_WordsInEnvironment;
		WordToLowestEditDistance_hmp_t	hmp_WordToLowestEditDistance;

		void IdentifyUnusedPotentialObjectWords (void);
		void IdentifyLowestEditDistances (void);


	public:
		WordObjects_vec_t				vec_WordObjects;
		EnhancedState*					p_EnhancedState;
		int*							p_UnusedPotentialObjectWords;
		int								i_UnusedPotentialObjectWordCount;

		RLState (void);
		~RLState (void);
		static void Init (void);

		static double NormalizedEditDistance (String _sLeft, String _sRight);
		static bool WordMaybeObject (String _sWord, String _sObjectName);
		static bool WordMaybeObjectStrict (String _sWord, String _sObjectName);
		bool CurrentStateHasWord (String& _sWord);
		double LowestEditDistanceInState (String& _sWord);

		EnvResponse_e ObserveState (EnhancedState* _pCurrentState);
		EnvResponse_e ObserveState (EnhancedState* _pCurrentState,
									ExperienceInterface* _pExperienceInterface,
									TextInterface* _pTextInterface);
		EnvResponseToAction_e ApplyAction (ExperienceInterface* _pExperienceInterface,
										   TextInterface* _pTextInterface,
										   RLAction& _rAction,
										   bool _bTrainingMode);

		bool WordLocationsAreValid (char _lWC, char _lWU, char _lWP)
			{ return p_EnhancedState->WordLocationsAreValid (_lWC, _lWU, _lWP); };

		String ConvertToCommandString (RLAction& _rAction);

		char TextLength (void)
			{ return p_EnhancedState->p_Document->i_WordCount; };
		int iGetWord (int _iLocation)
			{ return p_EnhancedState->p_Document->iGetWord (_iLocation); };
		String sGetWord (int _iLocation)
			{ return p_EnhancedState->p_Document->sGetWord (_iLocation); };
		int GetSentenceIndex (int _iLocation) const
			{
				if (-1 == _iLocation)
					return -1;
				return p_EnhancedState->p_Document->GetSentenceIndex (_iLocation);
			};

		char PreviousLocation_WU (void)
			{ return p_EnhancedState->i_PreviousLocation_WU; };
		char PreviousLocation_WC (void)
			{ return p_EnhancedState->i_PreviousLocation_WC; };
		char PreviousLocation_WP (void)
			{ return p_EnhancedState->i_PreviousLocation_WP; };
		char UnusedWordLocation (int _iIndex)
			{ return p_EnhancedState->p_UnusedWordLocations [_iIndex]; };
		char UnusedWordCount (void)
			{ return p_EnhancedState->i_UnusedWordLocationCount; };
		char* UnusedWordLocations (void)
			{ return p_EnhancedState->p_UnusedWordLocations; };

		EnvironmentState& GetEnvironmentState (void)
			{ return *(EnvironmentState*) p_EnhancedState; };
		int EnvironmentObjectCount (void)
			{ return p_EnhancedState->p_EnvironmentState->GetObjectCount (); };
		UIObject* GetObject (long _lIndex)
			{ return (UIObject*) p_EnhancedState->p_EnvironmentState->GetObject (_lIndex); };

		double GetReward (void)
			{ return p_EnhancedState->d_Reward; };
		void SetReward (double _dReward)
			{
				if (p_EnhancedState->d_Reward == _dReward)
					return;
				p_EnhancedState->d_Reward = _dReward;
			};
		void AddToEnvironmentStateReward (float _fReward)
			{ p_EnhancedState->p_EnvironmentState->AddReward (_fReward); };
		float GetAverageEnvironmentStateReward (void)
			{ return p_EnhancedState->p_EnvironmentState->GetAverageReward (); };

		bool IsStartState (void)
			{ return (NULL == p_EnhancedState->p_Parent); };

		bool IsActionWordless (RLAction& _rAction)
			{ return (_rAction.i_SelectedObjectWordLocation == TextLength ()); };
		int_set_t PotentialObjectWordsCoveredByState (char* _pWordLocations, int _iWordCount);
		int_set_t PotentialObjectWordsCoveredByStateStrict (char* _pWordLocations, int _iWordCount);
};



#endif
