#include <iostream>
#include <nlp_string.h>
#include <nlp_config.h>
#include <nlp_time.h>
#include <ili_cache.h>
#include "Learner.h"
#include "Vocabulary.h"
#include "Annotation.h"
#include "ReplayTrace.h"


extern void DumpSource (void);



double learning_rate (int _iTime, double _dBaseRate, double _c, double _tau)
{
	// search-then-converge learning rate schedule.
	double n0 = 1;

	double cn0 = _c/n0;
	double ttau = _iTime/_tau;

	return _dBaseRate * (1 + cn0 * ttau) / (1 + (cn0 + _iTime) * ttau);
}



int main (int argc, const char* argv [])
{
	Config::SetCommandLine (argc, argv);


	// copy config file to output path ...	
	{
		String sConfigFile (Config::GetConfigFileName ());
		String sTarget;
		sTarget << (config)"output_path" << '/';
		size_t iNameStart = sConfigFile.ReverseFind ("/");
		if (string::npos != iNameStart)
			sTarget << sConfigFile.substr (iNameStart);
		else
			sTarget << sConfigFile;
		Path::CopyFile (Config::GetConfigFileName (), sTarget);
	}

	// copy binary to output path...		
	{
		String sTarget;
		sTarget << (config)"output_path" << '/' << argv [0];
		Path::CopyFile (argv [0], sTarget);
	}


	// load data ...						
	Vocabulary::Load ();
	if (false == Document::Load ())
	{
		cout << "[ERROR]  Failed to load documents." << endl;
		return 1;
	}
	if (false == Annotation::Load ())
	{
		cout << "[ERROR]  Failed to load annotations." << endl;
		return 1;
	}
	if (false == ReplayTrace::Load ())
	{
		cout << "[ERROR]  Failed to load replay traces." << endl;
		return 1;
	}


	// initialize world interfaces ...		
    CacheInterface::i_ForegroundFeatureIndex = (config)"foreground_feature_index";
    CacheInterface::i_NewlyForegroundFeatureIndex = (config)"newly_foreground_feature_index";
    CacheInterface::b_Verbose = (1 == (int)(config)"verbose");
	EnvObject::i_ObjectMemorySize = sizeof (UIObject);


	// initialize learner...				
	Learner o_Learner;
	TextInterface			o_TextInterface;
	ExperienceInterface		o_ExperienceInterface;

	o_Learner.Init (&o_ExperienceInterface, &o_TextInterface);


	// initialize external interfaces...	
	if (false == o_ExperienceInterface.Start ())
	{
		cout << "[ERROR]  Failed to start environment interface." << endl;
		return 1;
	}
	if (1 == (int)(config)"known_world_only")
	{
		cout << "[INFO]  Switch to known-world-only mode? [y/n] : " << flush;
		char zInput [100];
		cin >> zInput;
		if (0 != strcmp ("y", zInput))
			return 1;

		if (false == o_ExperienceInterface.KnownWorldOnly ())
			return 1;
	}
	


	// learn ...							
	if (false == o_Learner.LoadWeights ())
		return 1;


	cout << "Starting learning iterations ..." << endl;
	long lMaxIterations = (config)"learning_iterations";
	long lTestInterval = (config)"test_interval";
	bool bForgetTestStateGraph = (1 == (int)(config)"forget_test_state_graph");

	int iConsequentTestRuns = (config)"consequent_test_runs";
	int iTrainingRepetitionsPerUpdate = (config)"training_repetitions_per_update";
	int iFirstStageRepetitions = (config)"first_stage_repetitions";
	int iSecondStageRepetitions = (config)"second_stage_repetitions";

	bool bUseA = (1 == (int)(config)"use_a");
	int iUseAFromIteration = (int)(config)"use_a_from_iteration";

	int iLookAtNextSentenceFromIteration = (int)(config)"look_at_next_sentence_from_iteration";
	EpisodeStep::b_LookAtNextSentenceAlso = false;

	double dBaseLearningRate = (config)"learning_rate";
	double dLearningRateC = (config)"learning_rate_c";
	double dLearningRateTau = (config)"learning_rate_tau";

	

	o_Learner.d_AlternateStartStateProbability = 0;
	for (long j = 0; j < iFirstStageRepetitions; ++ j)
		o_Learner.Learn (0, j);
	o_Learner.ForgetWeightUpdates ();

	for (long j = 0; j < iSecondStageRepetitions; ++ j)
		o_Learner.Learn (0, j);
	if (iSecondStageRepetitions > 0)
	{
		o_Learner.UpdateWeights ();
		o_Learner.SaveWeights ();
	}


	o_Learner.d_AlternateStartStateProbability = (config)"alternate_start_state_probability";

	int iTotalDocuments = Document::DocumentCount ();
	int iPositiveRewardDocuments = 0;
	for (long i = 1; i <= lMaxIterations; ++ i)
	{
		EpisodeStep::b_UseA = false;
		if ((true == bUseA) && (i >= iUseAFromIteration))
		{
			EpisodeStep::b_UseA = true;
			o_Learner.d_ExplorationRate = (config)"exploration_rate_with_a";
		}

		if (i >= iLookAtNextSentenceFromIteration)
			EpisodeStep::b_LookAtNextSentenceAlso = true;

		if (0 == (i % lTestInterval))
		{
			// we should use the lookahead feature	
			// always during test...				
			EpisodeStep::b_UseA = true;

			for (int j = 0; j < iConsequentTestRuns; ++ j)
				o_Learner.Test (i, j);
			if (true == bForgetTestStateGraph)
				Document::ForgetTestStateGraph ();
		}
		else
		{
			for (int j = 0; j < iTrainingRepetitionsPerUpdate; ++ j)
				iPositiveRewardDocuments = o_Learner.Learn (i, j);

			o_Learner.d_LearningRate = learning_rate (iPositiveRewardDocuments, 
													  dBaseLearningRate, 
													  dLearningRateC,
													  dLearningRateTau * iTotalDocuments);

			cout << " Exploration rate : " << o_Learner.d_ExplorationRate << endl;
			cout << " Learning rate : " << o_Learner.d_LearningRate 
				 << ", positives : " << iPositiveRewardDocuments 
				 << ", total : " << iTotalDocuments << endl;

			o_Learner.UpdateWeights ();
			o_Learner.SaveWeights ();
		}

		Cache* pCache = o_ExperienceInterface.GetCache ();
		size_t iCacheSize = pCache->GetNodeCount ();
		cout << " Cache size : " << iCacheSize << endl;
		if (iCacheSize > 36000)
			break;
	}


	cout << "Learning complete, shutting down ..." << endl;
	// shutdown world interfaces ...		
	o_ExperienceInterface.Stop ();
	cout << "   - done" << endl;
}
